from .questrade import Questrade, to_datestring
import datetime
from typing import Optional
from urllib.parse import urljoin
from .types import *


class Account(Questrade):
    """This class communicates with the Questrade Account APIs, in order to retrieve
    accunt-specific data.
    """

    def __init__(self, **kwargs):
        Questrade.__init__(self, **kwargs)

    def get_all(self, raw: Optional[bool]=False):
        """Returns the list of accounts associated.
        https://www.questrade.com/api/documentation/rest-operations/account-calls/accounts

        raw - If set, return the raw JSON rather than objects of qtype.
        """
        url = urljoin(self.server, "/v1/accounts")
        return self._request(url, qtype=TradingAccount, key="accounts", raw=raw)

    def positions(self, id: int, raw: Optional[bool]=False):
        """Return positions held by the specified account.
        https://www.questrade.com/api/documentation/rest-operations/account-calls/accounts-id-positions

        id - An integer containing the account ID.
        raw - If set, return the raw JSON rather than objects of qtype.
        """
        url = urljoin(self.server, "/v1/accounts/%d/positions" % id)
        return self._request(url, qtype=AccountPosition, key="positions", raw=raw)

    def activities(self, id: int, start: datetime.datetime, end: datetime.datetime, raw: Optional[bool]=False):
        """Return the account activities - actions including buys, sells, and dividends, amongst other things.
        https://www.questrade.com/api/documentation/rest-operations/account-calls/accounts-id-activities

        id - An integer containing the account ID.
        start - The start time of transactions
        end - The end time of the transactons
        raw - If set, return the raw JSON rather than objects of qtype.
        """
        if start > end:
            start_date = to_datestring(end)
            end_date = to_datestring(start)
        else:
            start_date = to_datestring(start)
            end_date = to_datestring(end)

        url = urljoin(self.server, "/v1/accounts/%d/activities/?startTime=%s&endTime=%s" % (id, start_date, end_date))
        return self._request(url, qtype=AccountActivity, key="activities", raw=raw)

    def executions(self, id: int, start: Optional[datetime.datetime]=None, end: Optional[datetime.datetime]=None, raw: Optional[bool]=False):
        """Return the account executions - actions including buys, sells, and dividends, amongst other things.
        https://www.questrade.com/api/documentation/rest-operations/account-calls/accounts-id-executions

        id - An integer containing the account ID.
        start - The start time of transactions
        end - The end time of the transactons
        raw - If set, return the raw JSON rather than objects of qtype.
        """

        if start is None:
            url = urljoin(self.server, "/v1/accounts/%d/executions" % id)
            return self._request(url, qtype=AccountExecution, key="executions", raw=raw)

        start_date, end_date = None, None
        if start > end:
            start_date = to_datestring(end)
            end_date = to_datestring(start)
        else:
            start_date = to_datestring(start)
            end_date = to_datestring(end)

        url = urljoin(self.server, "/v1/accounts/%d/executions?startTime=%s&endTime=%s" % (id, start_date, end_date))
        return self._request(url, qtype=AccountExecution, key="executions", raw=raw)

    def balances(self, id: int, raw: Optional[bool]=False):
        """Return the cash balances associated with the questrade account.
        https://www.questrade.com/api/documentation/rest-operations/account-calls/accounts-id-balances

        id - An integer containing the account ID.
        raw - If set, return the raw JSON rather than objects of qtype.
        """

        url = urljoin(self.server, "/v1/accounts/%d/balances" % id)
        return self._request(url, qtype=CurrencyBalance, key="perCurrencyBalances", raw=raw)

    def orders(self, id: int, start: Optional[datetime.datetime]=None, end: Optional[datetime.datetime]=None, state: Optional[str]='All', raw: Optional[bool]=False):
        """Return the account executions - actions including buys, sells, and dividends, amongst other things.
        https://www.questrade.com/api/documentation/rest-operations/account-calls/accounts-id-orders

        id - An integer containing the account ID.
        start - The start time of transactions
        end - The end time of the transactons
        raw - If set, return the raw JSON rather than objects of qtype.
        """

        valid_states = ['All', 'Open', 'Closed']
        if state not in valid_states:
            raise AttributeError("Invalid state. State must be one of %s" % valid_states)

        if start is None:
            url = urljoin(self.server, "/v1/accounts/%d/orders&stateFilter=%s" % (id, state))
            return self._request(url, qtype=Order, key="orders", raw=raw)

        start_date, end_date = None, None
        if start > end:
            start_date = to_datestring(end)
            end_date = to_datestring(start)
        else:
            start_date = to_datestring(start)
            end_date = to_datestring(end)

        url = urljoin(self.server, "/v1/accounts/%d/orders?startTime=%s&endTime=%s" % (id, start_date, end_date))
        return self._request(url, qtype=Order, key="orders", raw=raw)


import logging
from typing import Optional


class QuestradeType:
    """This class maps the questrade type data back to python objects. An active choice has been made to raise
    an AttributeError when the Questrade API no longer contains a field, or contains new fields.
    """

    LOGGER_NAME = "questrade-logger"

    def __init__(self, d: Optional[dict]=None, logger_name: Optional[str]=None):
        """Given that Questrade returns JSON items, all returns result in a dictionary.
        This function takes the dictionary and sets variables on the class object, for
        each key in the dictionary.
        """
        if logger_name is None:
            logger = logging.getLogger(self.LOGGER_NAME)
        else:
            logger = logging.getLogger(logger_name)
        if d is None:
            return

        for k, v in d.items():

            # sometimes fields are uppercase, sometimes they aren't
            if k.upper() not in self.fields:
                logger.warning("Invalid questrade field response. Missing %s. Received fields: %s" % (k, d.keys()))
            setattr(self, k.upper(), v)


class Auth(QuestradeType):
    """This type contains the results of a questrade auth request"""

    @property
    def fields(self):
        return [
            "ACCESS_TOKEN",
            "REFRESH_TOKEN",
            "EXPIRES_IN",
            "TOKEN_TYPE",
            "API_SERVER",
        ]


class OptionQuote(QuestradeType):
    """This type returns option types"""

    @property
    def fields(self):
        return [
            "UNDERLYING",
            "UNDERLYINGID",
            "SYMBOL",
            "SYMBOLID",
            "BIDPRICE",
            "BIDSIZE",
            "ASKPRICE",
            "ASKSIZE",
            "LASTTRADEPRICETRHRS",
            "LASTTRADEPRICE",
            "LASTTRADESIZE",
            "LASTTRADETICK",
            "LASTTRADETIME",
            "VOLUME",
            "OPENPRICE",
            "HIGHPRICEHIGHPRICE",
            "LOWPRICE",
            "VOLATILITY",
            "DELTA",
            "GAMMA",
            "THETA",
            "VEGA",
            "RHO",
            "OPENINTEREST",
            "DELAY",
            "ISHALTED",
            "VWAP",
        ]


class Quote(QuestradeType):
    """This type contains the results of a stock quote"""

    @property
    def fields(self):
        return [
            "SYMBOL",
            "SYMBOLID",
            "TIER",
            "BIDPRICE",
            "BIDSIZE",
            "ASKPRICE",
            "ASKSIZE",
            "LASTTRADEPRICETRHRS",
            "LASTTRADEPRICE",
            "LASTTRADESIZE",
            "LASTTRADETICK",
            "LASTTRADETIME",
            "VOLUME",
            "OPENPRICE",
            "HIGHPRICE",
            "LOWPRICE",
            "DELAY",
            "ISHALTED",
            "HIGH52W",
            "LOW52W",
            "VWAP",
        ]


class SearchSymbol(QuestradeType):
    """This type contains the results of a stock symbol search"""

    @property
    def fields(self):
        return [
            "SYMBOL",
            "SYMBOLID",
            "DESCRIPTION",
            "SECURITYTYPE",
            "LISTINGEXCHANGE",
            "ISTRADABLE",
            "ISQUOTABLE",
            "CURRENCY",
        ]


class SymbolData(QuestradeType):
    """This type contains the results of a stock symbol fetch"""

    @property
    def fields(self):
        return [
            "SYMBOL",
            "SYMBOLID",
            "PREVDAYCLOSEPRICE",
            "HIGHPRICE52",
            "LOWPRICE52",
            "AVERAGEVOL3MONTHS",
            "AVERAGEVOL20DAYS",
            "OUTSTANDINGSHARES",
            "EPS",
            "PE",
            "DIVIDEND",
            "YIELD",
            "EXDATE",
            "MARKETCAP",
            "TRADEUNIT",
            "OPTIONTYPE",
            "OPTIONDURATIONTYPE",
            "OPTIONROOT",
            "OPTIONCONTRACTDELIVERABLES",
            "OPTIONEXERCISETYPE",
            "LISTINGEXCHANGE",
            "DESCRIPTION",
            "SECURITYTYPE",
            "OPTIONEXPIRYDATE",
            "DIVIDENDDATE",
            "OPTIONSTRIKEPRICE",
            "ISTRADABLE",
            "ISQUOTABLE",
            "HASOPTIONS",
            "CURRENCY",
            "MINTICKS",
            "INDUSTRYSECTOR",
            "INDUSTRYGROUP",
            "INDUSTRYSUBGROUP",
        ]


class Candle(QuestradeType):
    """Historic end of day data"""

    @property
    def fields(self):
        return [
            "START",
            "END",
            "LOW",
            "HIGH",
            "OPEN",
            "CLOSE",
            "VOLUME",
            "VWAP",
        ]


class TradingAccount(QuestradeType):
    """Trading accounts, as held in questrade"""

    @property
    def fields(self):
        return [
            "TYPE",
            "NUMBER",
            "STATUS",
            "ISPRIMARY",
            "ISBILLING",
            "CLIENTACCOUNTTYPE"
        ]


class AccountPosition(QuestradeType):
    """The positions, i.e shares held by a trading account"""

    @property
    def fields(self):
        return [
            "SYMBOL",
            "SYMBOLID",
            "OPENQUANTITY",
            "CLOSEDQUANTITY",
            "CURRENTMARKETVALUE",
            "CURRENTPRICE",
            "AVERAGEENTRYPRICE",
            "DAYPNL",
            "CLOSEDPNL",
            "OPENPNL",
            "TOTALCOST",
            "ISREALTIME",
            "ISUNDERREORG"
        ]


class AccountActivity(QuestradeType):
    """Account activities - the actions (dividends, buy, sell, etc) that took place in a trading account"""

    @property
    def fields(self):
        return [
            "TRADEDATE",
            "TRANSACTIONDATE",
            "SETTLEMENTDATE",
            "ACTION",
            "SYMBOL",
            "SYMBOLID",
            "DESCRIPTION",
            "CURRENCY",
            "QUANTITY",
            "PRICE",
            "GROSSAMOUNT",
            "COMMISSION",
            "NETAMOUNT",
            "TYPE",
        ]


class AccountExecution(QuestradeType):
    """Executions - The individual actions that took place within an account, to close trades."""

    @property
    def fields(self):
        return [
            "SYMBOL",
            "SYMBOLID",
            "QUANTITY",
            "SIDE",
            "PRICE",
            "ID",
            "ORDERID",
            "ORDERCHAINID",
            "EXCHANGEEXECID",
            "TIMESTAMP",
            "NOTES",
            "VENUE",
            "TOTALCOST",
            "ORDERPLACEMENTCOMMISSION",
            "COMMISSION",
            "EXECUTIONFEE",
            "SECFEE",
            "LEGID",
            "CANADIANEXECUTIONFEE",
            "PARENTID",
        ]


class CurrencyBalance(QuestradeType):
    """CurrencyBalance - the object contained in all sub balance calls."""

    @property
    def fields(self):
        return [
            "CURRENCY",
            "CASH",
            "MARKETVALUE",
            "TOTALEQUITY",
            "BUYINGPOWER",
            "MAINTENANCEEXCESS",
            "ISREALTIME",
        ]


class Order(QuestradeType):
    """Order - An order created on an account."""

    @property
    def fields(self):
        return [
			"ID",
			"SYMBOL",
			"SYMBOLID",
			"TOTALQUANTITY",
			"OPENQUANTITY",
			"FILLEDQUANTITY",
			"CANCELEDQUANTITY",
			"SIDE",
			"TYPE",
			"LIMITPRICE",
			"STOPPRICE",
			"ISALLORNONE",
			"ISANONYMOUS",
			"ICEBERGQUANTITY",
			"MINQUANTITY",
			"AVGEXECPRICE",
			"LASTEXECPRICE",
			"SOURCE",
			"TIMEINFORCE",
			"GTDDATE",
			"STATE",
            "REJECTIONREASON"
			"CHAINID",
			"CREATIONTIME",
			"UPDATETIME",
			"NOTES",
			"PRIMARYROUTE",
			"SECONDARYROUTE",
			"ORDERROUTE",
			"VENUEHOLDINGORDER",
			"COMISSIONCHARGED",
			"EXCHANGEORDERID",
			"ISSIGNIFICANTSHAREHOLDER",
			"ISINSIDER",
			"ISLIMITOFFSETINDOLLAR",
			"USERID",
			"PLACEMENTCOMMISSION",
			"LEGS",
			"STRATEGYTYPE",
			"TRIGGERSTOPPRICE",
			"ORDERGROUPID",
			"ORDERCLASS",
            "ORDERTYPE",
            "ISCROSSZERO",
        ]


from .questrade import Questrade, to_datestring
import datetime
from typing import Optional
from urllib.parse import urljoin
from .types import Candle, Quote, SymbolData, SearchSymbol


class Symbol(Questrade):
    """This class communicates with the various symbol APIs in order to provide
    symbol data such as price history and current quotes.
    """

    def __init__(self, **kwargs):
        Questrade.__init__(self, **kwargs)

    def get(self, ids: Optional[int]=None, symbols: Optional[str]=None, raw: Optional[bool]=False):
        """Retrieve detailed information about one or more symbol.
        https://www.questrade.com/api/documentation/rest-operations/market-calls/symbols-id

        ids - A list of one or more questrade stock symbol ids.
        symbols - A list of one or more stock symbols.
        raw - If set, return the raw JSON rather than objects of qtype.
        """
        if not ids and not symbols or ids is not None and symbols is not None:
            raise AttributeError("either a list of ids or symbols must be specified")

        url = None
        if ids:
            qids = ','.join(str(i) for i in ids)
            url = urljoin(self.server, "/v1/symbols/?ids=%s" % qids)

        if symbols:
            qnames = ','.join(symbols)
            url = urljoin(self.server, "/v1/symbols/?names=%s" % qnames)

        return self._request(url, qtype=SymbolData, key="symbols", raw=raw)

    def search(self, sym: str, raw: Optional[bool]=False):
        """Search questrade for a matching stock symbol.
        https://www.questrade.com/api/documentation/rest-operations/market-calls/symbols-search

        sym: A string containing a stock symbol
        raw - If set, return the raw JSON rather than objects of qtype.
        """
        url = urljoin(self.server, "/v1/symbols/search?prefix=%s" % sym)
        return self._request(url, qtype=SearchSymbol, key="symbols", raw=raw)

    def quotes(self, ids: list[int], raw: Optional[bool]=False):
        """Retrieves the most recent quote data for a list of stock symbols.
        https://www.questrade.com/api/documentation/rest-operations/market-calls/markets-quotes-id

        ids - A list of questrade IDs whose stock quote data is to be retrieved.
        raw - If set, return the raw JSON rather than objects of qtype.
        """
        qids = ','.join(str(i) for i in ids)
        url = urljoin(self.server, "/v1/markets/quotes?ids=%s" % qids)
        return self._request(url, qtype=Quote, key="quotes", raw=raw)

    def history(self, id: int, start: datetime.datetime, end: datetime.datetime, interval: str="OneDay", raw: Optional[bool]=False):
        """Returns historical market data in an OHLC candlesick, for the provided symbol.
        https://www.questrade.com/api/documentation/rest-operations/market-calls/markets-candles-id

        id - An integer containing the internal questrade ID.
        start - The start time of the candle.
        end - The end time of the candle.
        interval - The interval for the candle data.
        raw - If set, return the raw JSON rather than objects of qtype.
        """
        if start > end:
            start_date = to_datestring(end)
            end_date = to_datestring(start)
        else:
            start_date = to_datestring(start)
            end_date = to_datestring(end)

        url = urljoin(self.server, "/v1/markets/candles/%s?startTime=%s&endTime=%s&interval=%s" % (id, start_date, end_date, interval))
        return self._request(url, qtype=Candle, key="candles", raw=raw)


from questradeist.types import QuestradeType
from .auth import QuestradeAuth
import datetime
import requests


def to_datetime(date):
    """A uniform function for parsing a questrade date and returning a datetime object."""
    return datetime.datetime.strptime(date, "%Y-%m-%dT00:00:00.000000-%H:%M")


def to_datestring(date):
    """A uniform function for formatting datetime objects like the strings Questrade expects."""
    return datetime.datetime.strftime(date, "%Y-%m-%dT00:00:00.00-05:00")


class Questrade(object):
    """This is the base call for all questrade operations."""

    def __init__(self, access_token: str=None, refresh_token: str=None, f: callable=None, expires: datetime.datetime=None):
        qtauth = QuestradeAuth(access_token, refresh_token, expires)
        self._setup(qtauth, f)

    def _setup(self, qtauth, f: callable=None):
        """Setup embeds the questrade AUTH api results into this class."""
        setattr(self, "ACCESS_TOKEN", qtauth.AUTH.ACCESS_TOKEN)
        setattr(self, "REFRESH_TOKEN", qtauth.AUTH.REFRESH_TOKEN)
        setattr(self, "API_SERVER", qtauth.AUTH.API_SERVER)
        setattr(self, "EXPIRES", qtauth.AUTH.EXPIRES)

        if f:
            f(qtauth.AUTH)

    @property
    def access_token(self):
        """Returns the access token"""
        return self.ACCESS_TOKEN

    @property
    def refresh_token(self):
        """Returns the refresh token"""
        return self.REFRESH_TOKEN

    @property
    def expires(self):
        """Returns the time the access token expires"""
        return self.EXPIRES

    @property
    def server(self):
        """Returns the API server to be used in Questrade calls"""
        return self.API_SERVER

    def time(self):
        """Call the Questrade time api, and return the time associated
        with the service.
        """
        url = "/v1/time"
        return self._request(url, qtype=None, raw=True)

    def _request(self, url: str, qtype: QuestradeType, key: str=None, raw: bool=False):
        """This is a request wrapper. It's meant to be called by other functions.
        qtype - The types.Questrade object that should be used for deserializing data.
        key - A string, containing the key within the Questrade API response, that contains the response elements.
        raw - If set, return the raw JSON rather than objects of qtype.
        """

        header = {"Authorization": "Bearer %s" % self.access_token}
        r = requests.get(url, headers=header)
        if r.status_code != 200:

            # at least try to trigger a refresh if authentication fails.
            if datetime.datetime.now() > self.expires:
                qtauth = QuestradeAuth(refresh_token=self.refresh_token)
                self._setup(qtauth)
                r = requests.get(url, headers=header)
                if r.status_code != 200:
                    raise OSError(r.content.decode('utf-8'))
            else:
                raise OSError(r.content.decode('utf-8'))

        if raw:
            return r.json()

        if key is not None:
            objs = r.json()[key]
        else:
            objs = r.json()

        qts = []
        for o in objs:
            qt = qtype(d=o)
            qts.append(qt)
        return qts



import requests
import datetime
from .types import Auth


class QuestradeAuth(object):
    """This class implements Questrade authentication.
    During the first time authorization, the access token must be exchanged
    for a full set of tokens. This means that this class must be initialized
    as QuestradeAuth(refresh_token=XXX) during first time initializtion.
    https://www.questrade.com/api/documentation/authorization
    """

    def __init__(self, access_token=None, refresh_token=None, expires=None):
        """Constructor
        access_token - The token used to access the Questrade API.
        refresh_token - Upon expiration of the access token, this token is used to trigger a refresh
        """
        if expires is not None and refresh_token is not None:
            if expires <= datetime.datetime.now():
                self.refresh(refresh_token)
                return

        if access_token is None and refresh_token is None:
            raise AttributeError("either access_token or refresh_token must be specified.")

        if access_token is not None:
            self.AUTH = Auth()
            self.AUTH.ACCESS_TOKEN = access_token
            return

        if refresh_token is not None:
            self.refresh(refresh_token)

    def refresh(self, refresh_token):
        """Refresh accesses the questrade API, refreshing the access_token
        to be used in API calls.
        refresh_token - The token to be used, exchanging for an access token.
        """

        authapi = "https://login.questrade.com/oauth2/token?grant_type=refresh_token&refresh_token=%s" % refresh_token
        now = datetime.datetime.now()
        r = requests.get(authapi)

        if r.status_code != 200:
            raise requests.exceptions.HTTPError(r.content)

        self.AUTH = Auth(r.json())

        # rather than constantly recalculate the EXPIRY period for the token
        # we do it once, here.
        self.AUTH.EXPIRES = now + datetime.timedelta(seconds=self.AUTH.EXPIRES_IN)
        return self.AUTH
